@@ -1,5 +1,5 @@
 /*
- * Copyright © 2019 – 2021 Red Hat Inc.
+ * Copyright © 2019 – 2023 Red Hat Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -31,7 +31,7 @@ import (
 	"github.com/godbus/dbus/v5"
 	"github.com/sirupsen/logrus"
 	"github.com/spf13/cobra"
-	"golang.org/x/crypto/ssh/terminal"
+	"golang.org/x/term"
 )
 
 const (
@@ -42,6 +42,7 @@ const (
 
 var (
 	createFlags struct {
+		authFile  string
 		container string
 		distro    string
 		image     string
@@ -58,14 +59,20 @@ var (
 )
 
 var createCmd = &cobra.Command{
-	Use:   "create",
-	Short: "Create a new toolbox container",
-	RunE:  create,
+	Use:               "create",
+	Short:             "Create a new toolbox container",
+	RunE:              create,
+	ValidArgsFunction: completionEmpty,
 }
 
 func init() {
 	flags := createCmd.Flags()
 
+	flags.StringVar(&createFlags.authFile,
+		"authfile",
+		"",
+		"Path to a file with credentials for authenticating to the registry for private images")
+
 	flags.StringVarP(&createFlags.container,
 		"container",
 		"c",
@@ -91,6 +98,17 @@ func init() {
 		"Create a toolbox container for a different operating system release than the host")
 
 	createCmd.SetHelpFunc(createHelp)
+
+	if err := createCmd.RegisterFlagCompletionFunc("distro", completionDistroNames); err != nil {
+		panicMsg := fmt.Sprintf("failed to register flag completion function: %v", err)
+		panic(panicMsg)
+	}
+
+	if err := createCmd.RegisterFlagCompletionFunc("image", completionImageNames); err != nil {
+		panicMsg := fmt.Sprintf("failed to register flag completion function: %v", err)
+		panic(panicMsg)
+	}
+
 	rootCmd.AddCommand(createCmd)
 }
 
@@ -108,11 +126,33 @@ func create(cmd *cobra.Command, args []string) error {
 	}
 
 	if cmd.Flag("distro").Changed && cmd.Flag("image").Changed {
-		return errors.New("options --distro and --image cannot be used together")
+		var builder strings.Builder
+		fmt.Fprintf(&builder, "options --distro and --image cannot be used together\n")
+		fmt.Fprintf(&builder, "Run '%s --help' for usage.", executableBase)
+
+		errMsg := builder.String()
+		return errors.New(errMsg)
 	}
 
 	if cmd.Flag("image").Changed && cmd.Flag("release").Changed {
-		return errors.New("options --image and --release cannot be used together")
+		var builder strings.Builder
+		fmt.Fprintf(&builder, "options --image and --release cannot be used together\n")
+		fmt.Fprintf(&builder, "Run '%s --help' for usage.", executableBase)
+
+		errMsg := builder.String()
+		return errors.New(errMsg)
+	}
+
+	if cmd.Flag("authfile").Changed {
+		if !utils.PathExists(createFlags.authFile) {
+			var builder strings.Builder
+			fmt.Fprintf(&builder, "file %s not found\n", createFlags.authFile)
+			fmt.Fprintf(&builder, "'podman login' can be used to create the file.\n")
+			fmt.Fprintf(&builder, "Run '%s --help' for usage.", executableBase)
+
+			errMsg := builder.String()
+			return errors.New(errMsg)
+		}
 	}
 
 	var container string
@@ -126,44 +166,24 @@ func create(cmd *cobra.Command, args []string) error {
 		containerArg = "--container"
 	}
 
-	if container != "" {
-		if !utils.IsContainerNameValid(container) {
-			var builder strings.Builder
-			fmt.Fprintf(&builder, "invalid argument for '%s'\n", containerArg)
-			fmt.Fprintf(&builder, "Container names must match '%s'\n", utils.ContainerNameRegexp)
-			fmt.Fprintf(&builder, "Run '%s --help' for usage.", executableBase)
-
-			errMsg := builder.String()
-			return errors.New(errMsg)
-		}
-	}
-
-	var release string
-	if createFlags.release != "" {
-		var err error
-		release, err = utils.ParseRelease(createFlags.distro, createFlags.release)
-		if err != nil {
-			err := utils.CreateErrorInvalidRelease(executableBase)
-			return err
-		}
-	}
-
-	container, image, release, err := utils.ResolveContainerAndImageNames(container,
+	container, image, release, err := resolveContainerAndImageNames(container,
+		containerArg,
 		createFlags.distro,
 		createFlags.image,
-		release)
+		createFlags.release)
+
 	if err != nil {
 		return err
 	}
 
-	if err := createContainer(container, image, release, true); err != nil {
+	if err := createContainer(container, image, release, createFlags.authFile, true); err != nil {
 		return err
 	}
 
 	return nil
 }
 
-func createContainer(container, image, release string, showCommandToEnter bool) error {
+func createContainer(container, image, release, authFile string, showCommandToEnter bool) error {
 	if container == "" {
 		panic("container not specified")
 	}
@@ -176,7 +196,7 @@ func createContainer(container, image, release string, showCommandToEnter bool)
 		panic("release not specified")
 	}
 
-	enterCommand := getEnterCommand(container, release)
+	enterCommand := getEnterCommand(container)
 
 	logrus.Debugf("Checking if container %s already exists", container)
 
@@ -190,7 +210,7 @@ func createContainer(container, image, release string, showCommandToEnter bool)
 		return errors.New(errMsg)
 	}
 
-	pulled, err := pullImage(image, release)
+	pulled, err := pullImage(image, release, authFile)
 	if err != nil {
 		return err
 	}
@@ -265,28 +285,6 @@ func createContainer(container, image, release string, showCommandToEnter bool)
 	logrus.Debugf("%s canonicalized to %s", currentUser.HomeDir, homeDirEvaled)
 	homeDirMountArg := homeDirEvaled + ":" + homeDirEvaled + ":rslave"
 
-	bootMountFlags := "ro"
-	isBootReadWrite, err := isPathReadWrite("/boot")
-	if err != nil {
-		return err
-	}
-	if isBootReadWrite {
-		bootMountFlags = "rw"
-	}
-
-	bootMountArg := "/boot:/run/host/boot:" + bootMountFlags + ",rslave"
-
-	usrMountFlags := "ro"
-	isUsrReadWrite, err := isPathReadWrite("/usr")
-	if err != nil {
-		return err
-	}
-	if isUsrReadWrite {
-		usrMountFlags = "rw"
-	}
-
-	usrMountArg := "/usr:/run/host/usr:" + usrMountFlags + ",rslave"
-
 	var avahiSocketMount []string
 
 	avahiSocket, err := getServiceSocket("Avahi", "avahi-daemon.socket")
@@ -309,6 +307,17 @@ func createContainer(container, image, release string, showCommandToEnter bool)
 		kcmSocketMount = []string{"--volume", kcmSocketMountArg}
 	}
 
+	var pcscSocketMount []string
+
+	pcscSocket, err := getServiceSocket("pcsc", "pcscd.socket")
+	if err != nil {
+		logrus.Debug(err)
+	}
+	if pcscSocket != "" {
+		pcscSocketMountArg := pcscSocket + ":" + pcscSocket
+		pcscSocketMount = []string{"--volume", pcscSocketMountArg}
+	}
+
 	var mediaLink []string
 	var mediaMount []string
 
@@ -394,6 +403,7 @@ func createContainer(container, image, release string, showCommandToEnter bool)
 	createArgs := []string{
 		"--log-level", logLevelString,
 		"create",
+		"--cgroupns", "host",
 		"--dns", "none",
 		"--env", toolboxPathEnvArg,
 	}
@@ -404,7 +414,6 @@ func createContainer(container, image, release string, showCommandToEnter bool)
 		"--hostname", "toolbox",
 		"--ipc", "host",
 		"--label", "com.github.containers.toolbox=true",
-		"--label", "com.github.debarshiray.toolbox=true",
 	}...)
 
 	createArgs = append(createArgs, devPtsMount...)
@@ -423,16 +432,11 @@ func createContainer(container, image, release string, showCommandToEnter bool)
 	createArgs = append(createArgs, []string{
 		"--userns", usernsArg,
 		"--user", "root:root",
-		"--volume", "/etc:/run/host/etc",
+		"--volume", "/:/run/host:rslave",
 		"--volume", "/dev:/dev:rslave",
-		"--volume", "/run:/run/host/run:rslave",
-		"--volume", "/tmp:/run/host/tmp:rslave",
-		"--volume", "/var:/run/host/var:rslave",
-		"--volume", bootMountArg,
 		"--volume", dbusSystemSocketMountArg,
 		"--volume", homeDirMountArg,
 		"--volume", toolboxPathMountArg,
-		"--volume", usrMountArg,
 		"--volume", runtimeDirectoryMountArg,
 	}...)
 
@@ -440,6 +444,7 @@ func createContainer(container, image, release string, showCommandToEnter bool)
 	createArgs = append(createArgs, kcmSocketMount...)
 	createArgs = append(createArgs, mediaMount...)
 	createArgs = append(createArgs, mntMount...)
+	createArgs = append(createArgs, pcscSocketMount...)
 	createArgs = append(createArgs, runMediaMount...)
 	createArgs = append(createArgs, toolboxShMount...)
 
@@ -459,7 +464,7 @@ func createContainer(container, image, release string, showCommandToEnter bool)
 
 	stdoutFd := os.Stdout.Fd()
 	stdoutFdInt := int(stdoutFd)
-	if logLevel := logrus.GetLevel(); logLevel < logrus.DebugLevel && terminal.IsTerminal(stdoutFdInt) {
+	if logLevel := logrus.GetLevel(); logLevel < logrus.DebugLevel && term.IsTerminal(stdoutFdInt) {
 		s.Prefix = fmt.Sprintf("Creating container %s: ", container)
 		s.Writer = os.Stdout
 		s.Start()
@@ -470,7 +475,7 @@ func createContainer(container, image, release string, showCommandToEnter bool)
 		return fmt.Errorf("failed to create container %s", container)
 	}
 
-	// The spinner must be stopped before showing the 'enter' hit below.
+	// The spinner must be stopped before showing the 'enter' hint below.
 	s.Stop()
 
 	if showCommandToEnter {
@@ -496,7 +501,7 @@ func createHelp(cmd *cobra.Command, args []string) {
 		return
 	}
 
-	if err := utils.ShowManual("toolbox-create"); err != nil {
+	if err := showManual("toolbox-create"); err != nil {
 		fmt.Fprintf(os.Stderr, "Error: %s\n", err)
 		return
 	}
@@ -528,15 +533,12 @@ func getDBusSystemSocket() (string, error) {
 	return pathEvaled, nil
 }
 
-func getEnterCommand(container, release string) string {
+func getEnterCommand(container string) string {
 	var enterCommand string
-	containerNamePrefixDefaultWithRelease := utils.ContainerNamePrefixDefault + "-" + release
 
 	switch container {
 	case utils.ContainerNameDefault:
 		enterCommand = fmt.Sprintf("%s enter", executableBase)
-	case containerNamePrefixDefaultWithRelease:
-		enterCommand = fmt.Sprintf("%s enter --release %s", executableBase, release)
 	default:
 		enterCommand = fmt.Sprintf("%s enter %s", executableBase, container)
 	}
@@ -644,32 +646,8 @@ func getServiceSocket(serviceName string, unitName string) (string, error) {
 	return "", fmt.Errorf("failed to find a SOCK_STREAM socket for %s", unitName)
 }
 
-func isPathReadWrite(path string) (bool, error) {
-	logrus.Debugf("Checking if %s is mounted read-only or read-write", path)
-
-	mountPoint, err := utils.GetMountPoint(path)
-	if err != nil {
-		return false, fmt.Errorf("failed to get the mount-point of %s: %s", path, err)
-	}
-
-	logrus.Debugf("Mount-point of %s is %s", path, mountPoint)
-
-	mountFlags, err := utils.GetMountOptions(mountPoint)
-	if err != nil {
-		return false, fmt.Errorf("failed to get the mount options of %s: %s", mountPoint, err)
-	}
-
-	logrus.Debugf("Mount flags of %s on the host are %s", path, mountFlags)
-
-	if !strings.Contains(mountFlags, "ro") {
-		return true, nil
-	}
-
-	return false, nil
-}
-
-func pullImage(image, release string) (bool, error) {
-	if _, err := utils.ImageReferenceCanBeID(image); err == nil {
+func pullImage(image, release, authFile string) (bool, error) {
+	if ok := utils.ImageReferenceCanBeID(image); ok {
 		logrus.Debugf("Looking for image %s", image)
 
 		if _, err := podman.ImageExists(image); err == nil {
@@ -724,7 +702,7 @@ func pullImage(image, release string) (bool, error) {
 		fmt.Println("Image required to create toolbox container.")
 
 		prompt := fmt.Sprintf("Download %s (500MB)? [y/N]:", imageFull)
-		shouldPullImage = utils.AskForConfirmation(prompt)
+		shouldPullImage = askForConfirmation(prompt)
 	}
 
 	if !shouldPullImage {
@@ -735,7 +713,7 @@ func pullImage(image, release string) (bool, error) {
 
 	stdoutFd := os.Stdout.Fd()
 	stdoutFdInt := int(stdoutFd)
-	if logLevel := logrus.GetLevel(); logLevel < logrus.DebugLevel && terminal.IsTerminal(stdoutFdInt) {
+	if logLevel := logrus.GetLevel(); logLevel < logrus.DebugLevel && term.IsTerminal(stdoutFdInt) {
 		s := spinner.New(spinner.CharSets[9], 500*time.Millisecond)
 		s.Prefix = fmt.Sprintf("Pulling %s: ", imageFull)
 		s.Writer = os.Stdout
@@ -743,8 +721,14 @@ func pullImage(image, release string) (bool, error) {
 		defer s.Stop()
 	}
 
-	if err := podman.Pull(imageFull); err != nil {
-		return false, fmt.Errorf("failed to pull image %s", imageFull)
+	if err := podman.Pull(imageFull, authFile); err != nil {
+		var builder strings.Builder
+		fmt.Fprintf(&builder, "failed to pull image %s\n", imageFull)
+		fmt.Fprintf(&builder, "If it was a private image, log in with: podman login %s\n", domain)
+		fmt.Fprintf(&builder, "Use '%s --verbose ...' for further details.", executableBase)
+
+		errMsg := builder.String()
+		return false, errors.New(errMsg)
 	}
 
 	return true, nil
