@@ -1,5 +1,5 @@
 /*
- * Copyright © 2019 – 2021 Red Hat Inc.
+ * Copyright © 2019 – 2023 Red Hat Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -56,7 +56,11 @@ var (
 		{"/run/libvirt", "/run/host/run/libvirt", ""},
 		{"/run/systemd/journal", "/run/host/run/systemd/journal", ""},
 		{"/run/systemd/resolve", "/run/host/run/systemd/resolve", ""},
+		{"/run/systemd/sessions", "/run/host/run/systemd/sessions", ""},
+		{"/run/systemd/system", "/run/host/run/systemd/system", ""},
+		{"/run/systemd/users", "/run/host/run/systemd/users", ""},
 		{"/run/udev/data", "/run/host/run/udev/data", ""},
+		{"/run/udev/tags", "/run/host/run/udev/tags", ""},
 		{"/tmp", "/run/host/tmp", "rslave"},
 		{"/var/lib/flatpak", "/run/host/var/lib/flatpak", "ro"},
 		{"/var/lib/libvirt", "/run/host/var/lib/libvirt", ""},
@@ -85,7 +89,9 @@ func init() {
 		"home",
 		"",
 		"Create a user inside the toolbox container whose login directory is HOME")
-	initContainerCmd.MarkFlagRequired("home")
+	if err := initContainerCmd.MarkFlagRequired("home"); err != nil {
+		panic("Could not mark flag --home as required")
+	}
 
 	flags.BoolVar(&initContainerFlags.homeLink,
 		"home-link",
@@ -108,19 +114,25 @@ func init() {
 		"shell",
 		"",
 		"Create a user inside the toolbox container whose login shell is SHELL")
-	initContainerCmd.MarkFlagRequired("shell")
+	if err := initContainerCmd.MarkFlagRequired("shell"); err != nil {
+		panic("Could not mark flag --shell as required")
+	}
 
 	flags.IntVar(&initContainerFlags.uid,
 		"uid",
 		0,
 		"Create a user inside the toolbox container whose numerical user ID is UID")
-	initContainerCmd.MarkFlagRequired("uid")
+	if err := initContainerCmd.MarkFlagRequired("uid"); err != nil {
+		panic("Could not mark flag --uid as required")
+	}
 
 	flags.StringVar(&initContainerFlags.user,
 		"user",
 		"",
 		"Create a user inside the toolbox container whose login name is USER")
-	initContainerCmd.MarkFlagRequired("user")
+	if err := initContainerCmd.MarkFlagRequired("user"); err != nil {
+		panic("Could not mark flag --user as required")
+	}
 
 	initContainerCmd.SetHelpFunc(initContainerHelp)
 	rootCmd.AddCommand(initContainerCmd)
@@ -260,7 +272,25 @@ func initContainer(cmd *cobra.Command, args []string) error {
 		if err := ioutil.WriteFile("/etc/krb5.conf.d/kcm_default_ccache",
 			kcmConfigBytes,
 			0644); err != nil {
-			return errors.New("failed to set KCM as the defult Kerberos credential cache")
+			return errors.New("failed to set KCM as the default Kerberos credential cache")
+		}
+	}
+
+	if utils.PathExists("/usr/lib/rpm/macros.d") {
+		logrus.Debug("Configuring RPM to ignore bind mounts")
+
+		var builder strings.Builder
+		fmt.Fprintf(&builder, "# Written by Toolbox\n")
+		fmt.Fprintf(&builder, "# https://github.com/containers/toolbox\n")
+		fmt.Fprintf(&builder, "\n")
+		fmt.Fprintf(&builder, "%%_netsharedpath /dev:/media:/mnt:/proc:/sys:/tmp:/var/lib/flatpak:/var/lib/libvirt\n")
+
+		rpmConfigString := builder.String()
+		rpmConfigBytes := []byte(rpmConfigString)
+		if err := ioutil.WriteFile("/usr/lib/rpm/macros.d/macros.toolbox",
+			rpmConfigBytes,
+			0644); err != nil {
+			return fmt.Errorf("failed to configure RPM to ignore bind mounts: %w", err)
 		}
 	}
 
@@ -293,7 +323,7 @@ func initContainer(cmd *cobra.Command, args []string) error {
 	uidString := strconv.Itoa(initContainerFlags.uid)
 	targetUser, err := user.LookupId(uidString)
 	if err != nil {
-		return fmt.Errorf("failed to lookup user ID %s: %w", uidString, err)
+		return fmt.Errorf("failed to look up user ID %s: %w", uidString, err)
 	}
 
 	toolboxRuntimeDirectory, err := utils.GetRuntimeDirectory(targetUser)
@@ -332,7 +362,7 @@ func initContainer(cmd *cobra.Command, args []string) error {
 		}
 	}
 
-	return nil
+	// code should not be reached
 }
 
 func initContainerHelp(cmd *cobra.Command, args []string) {
@@ -350,7 +380,7 @@ func initContainerHelp(cmd *cobra.Command, args []string) {
 		return
 	}
 
-	if err := utils.ShowManual("toolbox-init-container"); err != nil {
+	if err := showManual("toolbox-init-container"); err != nil {
 		fmt.Fprintf(os.Stderr, "Error: %s\n", err)
 		return
 	}
@@ -388,7 +418,7 @@ func configureUsers(targetUserUid int,
 		}
 
 		if err := shell.Run("usermod", nil, nil, nil, usermodArgs...); err != nil {
-			return fmt.Errorf("failed to modify user %s with UID %d", targetUser, targetUserUid)
+			return fmt.Errorf("failed to modify user %s with UID %d: %w", targetUser, targetUserUid, err)
 		}
 	} else {
 		logrus.Debugf("Adding user %s with UID %d:", targetUser, targetUserUid)
@@ -408,20 +438,20 @@ func configureUsers(targetUserUid int,
 		}
 
 		if err := shell.Run("useradd", nil, nil, nil, useraddArgs...); err != nil {
-			return fmt.Errorf("failed to add user %s with UID %d", targetUser, targetUserUid)
+			return fmt.Errorf("failed to add user %s with UID %d: %w", targetUser, targetUserUid, err)
 		}
 	}
 
 	logrus.Debugf("Removing password for user %s", targetUser)
 
 	if err := shell.Run("passwd", nil, nil, nil, "--delete", targetUser); err != nil {
-		return fmt.Errorf("failed to remove password for user %s", targetUser)
+		return fmt.Errorf("failed to remove password for user %s: %w", targetUser, err)
 	}
 
 	logrus.Debug("Removing password for user root")
 
 	if err := shell.Run("passwd", nil, nil, nil, "--delete", "root"); err != nil {
-		return errors.New("failed to remove password for root")
+		return fmt.Errorf("failed to remove password for root: %w", err)
 	}
 
 	return nil
@@ -502,24 +532,22 @@ func mountBind(containerPath, source, flags string) error {
 // If it's not, then redirectPath will still proceed with the linking in two
 // different ways depending whether target is an absolute or a relative link:
 //
-//   * absolute - target's destination will be prepended with /run/host, and
-//                containerPath will be linked to the resulting path. This is
-//                an attempt to unbreak things, but if it doesn't work then
-//                it's the user's responsibility to fix it up.
+//   - absolute - target's destination will be prepended with /run/host, and
+//     containerPath will be linked to the resulting path. This is an attempt
+//     to unbreak things, but if it doesn't work then it's the user's
+//     responsibility to fix it up.
 //
-//                This is meant to address the common case where
-//                /etc/resolv.conf on the host (ie., /run/host/etc/resolv.conf
-//                inside the container) is an absolute symbolic link to
-//                something like /run/systemd/resolve/stub-resolv.conf. The
-//                container's /etc/resolv.conf will then get linked to
-//                /run/host/run/systemd/resolved/resolv-stub.conf.
+//     This is meant to address the common case where /etc/resolv.conf on the
+//     host (ie., /run/host/etc/resolv.conf inside the container) is an
+//     absolute symbolic link to /run/systemd/resolve/stub-resolv.conf. The
+//     container's /etc/resolv.conf will then get linked to
+//     /run/host/run/systemd/resolved/resolv-stub.conf.
 //
-//                This is why properly configured hosts should use relative
-//                symbolic links, because they don't need to be adjusted in
-//                such scenarios.
+//     This is why properly configured hosts should use relative symbolic
+//     links, because they don't need to be adjusted in such scenarios.
 //
-//   * relative - containerPath will be linked to the invalid target, and it's
-//                the user's responsibility to fix it up.
+//   - relative - containerPath will be linked to the invalid target, and it's
+//     the user's responsibility to fix it up.
 //
 // folder signifies if the target is a folder
 func redirectPath(containerPath, target string, folder bool) error {
